#!/usr/bin/perl
#
#  XPM-to-RLE compressor tool for Bitmap Projector
#  Copyright (C) 2011 Matt Evans
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Image is written out in horiz strips of vert bytes, from top left
# to top right, downwards.
# Read in an xpm, easiest to parse, and RLE encode to an output
# array.
# May be beneficial to just compress 0 and ff bytes.

use File::Basename;

if ($#ARGV != 1) {
    print "thing.pl <infile> <outfile>\n";
    exit(1);
}
my $infile = $ARGV[0];
my $outfile = $ARGV[1];

open(IF, "< $infile") || die("Can't open $infile");
open(OF, "> $outfile") || die("Can't open $outfile");

my @pixels = ();
my $height = 0;
my $width = 0;
my $state = 0;
my $name = basename($outfile);
$name =~ s/\..*$//;
my $curbyte = 0;
my $curcol = 0;

while (<IF>) {
    if ($state == 0) {
        if (/^"([0-9]+)\s([0-9]+)\s([0-9]+)\s([0-9]+)\s*",$/) {
            $width = $1;
            $height = $2;
            $numcols = $3;

            print "Image $width x $height, $numcols colours\n";
            if (($width & 7) != 0) {
                die "Image width not /8\n";
            }
            $state = 1;
        }
    } elsif ($state == 1) {
        if (/^\"(.) c (.*)\",/) {
            $cols{$1} = $curcol;
            $curcol++;
        } elsif (/pixels/) {
            print "Read $curcol colours\n";
            $state = 2;
        }
    } else {
        # Read pixels.  Top bits of a byte are leftmost, so read L->R
        # and shift left
        if (/^\"(.*)\"[,]*$/) {
            my $line = $1;
            foreach my $c (split(//, $line)) {
                my $pixvalue = $cols{$c};
		# Nokia screen seems to have 1 = black, 0 = white. Hm.
		# So, invert:
		$pixvalue ^= 1;

                $curbyte <<= 1;
                $curbyte |= $pixvalue;
                $x++;
                if (($x & 7) == 0) {
		    push(@pixels, ($curbyte));
                    $curbyte = 0;
                }
            }
            $x = 0;
            $y++;

            if ($y == $height) {
		break;
            }
        }
    }
}

print "Read ".($#pixels+1)."\n";
my @rotdata = ();

# OK, now pixel data is in horiz-normal-direction rows in @pixels.
# Scan through this to assemble rotated bytes, then RLE this.
for ($y = 0; $y < $height; $y += 8) {
    for ($x = 0; $x < $width; $x += 8) {
	my $a = $pixels[($x/8) + (($y + 0) * $width / 8)];
	my $b = $pixels[($x/8) + (($y + 1) * $width / 8)];
	my $c = $pixels[($x/8) + (($y + 2) * $width / 8)];
	my $d = $pixels[($x/8) + (($y + 3) * $width / 8)];
	my $e = $pixels[($x/8) + (($y + 4) * $width / 8)];
	my $f = $pixels[($x/8) + (($y + 5) * $width / 8)];
	my $g = $pixels[($x/8) + (($y + 6) * $width / 8)];
	my $h = $pixels[($x/8) + (($y + 7) * $width / 8)];
	my $loop;
	for ($loop = 0; $loop < 8; $loop++) {
	    push(@rotdata, (
		     (($a >> 7) & 0x01) |
		     (($b >> 6) & 0x02) |
		     (($c >> 5) & 0x04) |
		     (($d >> 4) & 0x08) |
		     (($e >> 3) & 0x10) |
		     (($f >> 2) & 0x20) |
		     (($g >> 1) & 0x40) |
		     (($h >> 0) & 0x80) ));
	    $a <<= 1;
	    $b <<= 1;
	    $c <<= 1;
	    $d <<= 1;
	    $e <<= 1;
	    $f <<= 1;
	    $g <<= 1;
	    $h <<= 1;
	}
    }
}

# Now RLE compress.
# Current way:
# 	[00][xx]: xx+1 zeroes
#	[ff][xx]: xx+1 ffs
#	[zz]	: Anything else passed through verbatim
#
# 'Proper' way:
# [tok][0xxxxxxx][...]: Token pair followed by x bytes of straight-through data
# [tok][1xxxxxxx][yy]:  Expanded into x bytes of y.
# Choose tok as an 'uncommon' byte in the input stream.
# For these purposes 0 or ff will be quite common... do the first.

my @rle = ();
my $rlestate = 0; # 0 = normal. 1 = on span (see rlebyte)
my $rlebyte = 0;
my $bin = 0;
my $rlecount = 0;
my $lastbyte;

# 0 0 0 1
# f 1 2 3 0 1

$lastbyte = $rotdata[0];
if ($lastbyte != 0 && $lastbyte != 0xff) {
    push(@rle, ($lastbyte));
} else {
    $rlecount = 1;
}

for ($bin = 1; $bin <= $#rotdata; $bin++) {
    my $p = $rotdata[$bin];

    if ($p != $lastbyte || (($p != 0) && ($p != 0xff))) {
	if ($rlecount != 0) {
	    # End of the run.
	    print "[$rlecount x $lastbyte]";
	    push(@rle, ($lastbyte, $rlecount-1));
	    $rlecount = 0;
	}
	if (($p != 0) && ($p != 0xff)) {
	    print "($p)";
	    # Just pass through non-plain bytes.
	    push(@rle, ($p));
	} else {
	    # Saw 0/0xff, start new run
	    $rlecount = 1;
	}
    } else { # Same as last time and a 0/0xff
	# Byte same as last time
	$rlecount++;
	if ($rlecount == 256) {
	    # Got to maximum span, output it.
	    print "[$rlecount x $lastbyte]";
	    push(@rle, ($lastbyte, $rlecount-1));
	    $rlecount = 0;
	}
    }
    $lastbyte = $p;
}
# Close any open span
if ($rlecount != 0) {
    print "[$rlecount x $lastbyte]";
    push(@rle, ($lastbyte, $rlecount-1));
}

print "\nRLE len ".($#rle+1)."\n";

print OF "#define ".$name."_width $width\n";
print OF "#define ".$name."_height $height\n";
print OF "#define ".$name."_len ".($#rle+1)."\n";
print OF "PROGMEM char ".$name."[".($#rle + 1)."] = {\n";
for ($i = 0; $i <= $#rle; $i++) {
    print OF $rle[$i].", ";
}
print OF "\n};\n";

